/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###MODULE###
Name:      Strings
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 1998-2004.  All rights reserved.
Date:      January 2004
Version:   1.0
Status:    Stable

###DESCRIPTION###
<p>The Strings module provides a variety of string functions that are common
to most programming languages and operating systems. This includes support for
string conversions, searching, merging and so forth.  This module supports
strings of an unlimited length, so you do not need to worry about any possible
limitations.  As a bonus, this means that you can also use the string
functions on large text files and documents (so long as you null terminate
them).</p>

<p>The most important thing to remember is that some functions will return
strings that have had their space allocated with the
<a href="kernel.html#AllocMemory()">AllocMemory()</a> function.  If this
is the case for a function you are using, make sure you call <a
href="kernel.html#FreeMemory()">FreeMemory()</a> on the resulting
string before your program exits.</p>

###END###

*****************************************************************************/

#define LOCAL_STRINGS
#include <stdio.h>
#include <pandora/system/all.h>
#include <pandora/modules/strings.h>
#include <pandora/misc/time.h>

ERROR CMDInit(OBJECTPTR, struct KernelBase *);
ERROR CMDOpen(OBJECTPTR);

static LONG   FloatToStr(DOUBLE, STRING, LONG);
static LONG   IntToStr(LARGE, STRING, LONG);
static LONG   CharCopy(STRING, STRING, LONG);
static STRING * StrBuildArray(STRING, LONG, LONG, LONG);
static ERROR  StrCalculate(STRING, DOUBLE *, STRING, LONG);
static void   StrCapitalise(STRING);
static STRING StrClone(STRING);
static LONG   StrCopy(STRING, STRING, LONG);
static LONG   StrDatatype(STRING);
static LONG   StrExpand(STRING, LONG, LONG);
static ERROR  StrFormat(STRING, LONG, STRING, STRING, ...);
static ERROR  StrFormatDate(STRING, LONG, STRING, struct Time *);
static ERROR  StrInsert(STRING, STRING, LONG, LONG, LONG);
static LONG   StrLength(STRING);
static LONG   StrLineLength(STRING);
static void   StrLower(STRING);
static STRING StrNextLine(STRING);
static ERROR  StrReadColour(STRING, LONG *, LONG *, LONG *, LONG *);
static ERROR  StrReplace(STRING, STRING, STRING, STRING *, LONG);
static LONG   StrSearch(STRING, STRING, LONG);
static LONG   StrShrink(STRING, LONG, LONG);
static ERROR  StrSort(STRING *, LONG);
static DOUBLE StrToFloat(STRING);
static LARGE  StrToInt(STRING);
static ERROR  StrTranslate(STRING, LONG, LONG);
static void   StrUpper(STRING);
static LONG   UTF8CharacterLength(STRING);
static LONG   UTF8CharacterOffset(STRING, LONG);
static LONG   UTF8Length(STRING);
static LONG   UTF8OffsetToCharacter(STRING, LONG);
static LONG   UTF8PrevLength(STRING, LONG);
static LONG   UTF8ReadValue(STRING, LONG *);
static LONG   UTF8WriteValue(LONG, STRING, LONG);

static struct FunctionField argsCharCopy[]              = { { "TotalChars",   ARG_LONG },   { "String",  ARG_STRING }, { "Destination", ARG_STRING }, { "Length", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsFloatToStr[]            = { { "TotalChars",   ARG_LONG },   { "Float",   ARG_DOUBLE }, { "String", ARG_STRING }, { "StringSize", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsIntToStr[]              = { { "TotalChars",   ARG_LONG },   { "Integer", ARG_LARGE }, { "String", ARG_STRING }, { "StringSize", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrBuildArray[]         = { { "Array",        ARG_PTR },    { "List",    ARG_PTR }, { "Size", ARG_LONG }, { "AmtEntries", ARG_LONG }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrCalculate[]          = { { "Error",        ARG_ERROR },  { "String",  ARG_STRING }, { "Result", ARG_DOUBLE|ARG_PTR }, { "Buffer", ARG_STRING }, { "BufferSize", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrCapitalise[]         = { { "Void",         NULL },       { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrClone[]              = { { "ClonedString", ARG_STRING }, { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrCopy[]               = { { "TotalChars",   ARG_LONG },   { "String",  ARG_STRING }, { "Destination", ARG_STRING }, { "Length", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrDatatype[]           = { { "Datatype",     ARG_LONG },   { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrExpand[]             = { { "NewLength",    ARG_LONG },   { "String",  ARG_STRING }, { "Position", ARG_LONG }, { "TotalChars", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrFormat[]             = { { "Error",        ARG_LONG },   { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrFormatDate[]         = { { "Error",        ARG_LONG },   { "Buffer",  ARG_STRING }, { "BufferSize", ARG_LONG }, { "Format", ARG_STRING }, { "Time", ARG_PTR }, { NULL, NULL } };
static struct FunctionField argsStrInsert[]             = { { "Error",        ARG_LONG },   { "Insert",  ARG_STRING }, { "Buffer", ARG_STRING }, { "Position", ARG_LONG }, { "ReplaceChars", ARG_LONG }, { "BufferSize", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrLength[]             = { { "Length",       ARG_LONG },   { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrLineLength[]         = { { "Length",       ARG_LONG },   { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrLower[]              = { { "Void",         NULL },       { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrNextLine[]           = { { "String",       ARG_STRING }, { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrReadColour[]         = { { "Error",        ARG_ERROR },  { "Colour",  ARG_STRING }, { "Red", ARG_LONG|ARG_PTR }, { "Green", ARG_LONG|ARG_PTR }, { "Blue", ARG_LONG|ARG_PTR }, { "Alpha", ARG_LONG|ARG_PTR }, { NULL, NULL } };
static struct FunctionField argsStrReplace[]            = { { "Error",        ARG_ERROR },  { "String",  ARG_STRING }, { "Keyword", ARG_STRING }, { "Replacement", ARG_STRING }, { "Result", ARG_STRING }, { "CaseSensitive", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrSearch[]             = { { "Position",     ARG_LONG },   { "Keyword", ARG_STRING }, { "String", ARG_STRING }, { "CaseSensitive", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrShrink[]             = { { "NewLength",    ARG_LONG },   { "String",  ARG_STRING }, { "Position", ARG_LONG }, { "TotalChars", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrSort[]               = { { "Error",        ARG_ERROR },  { "Array",   ARG_PTR }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrToFloat[]            = { { "Float",        ARG_DOUBLE }, { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrToInt[]              = { { "Integer",      ARG_LARGE },  { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsStrTranslate[]          = { { "Error",        ARG_ERROR },  { "Buffer",  ARG_STRING }, { "BufferSize", ARG_LONG }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsStrUpper[]              = { { "Void",         NULL },       { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsUTF8CharacterOffset[]   = { { "Offset",       ARG_LONG },   { "String",  ARG_STRING }, { "Index", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsUTF8Length[]            = { { "Length",       ARG_LONG },   { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsUTF8OffsetToCharacter[] = { { "Result",       ARG_LONG },   { "String",  ARG_STRING }, { "Offset", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsUTF8PrevLength[]        = { { "Length",       ARG_LONG },   { "String",  ARG_STRING }, { "Index", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsUTF8CharacterLength[]   = { { "Length",       ARG_LONG },   { "String",  ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsUTF8ReadValue[]         = { { "Value",        ARG_LONG },   { "String",  ARG_STRING }, { "Length", ARG_LONG|ARG_PTR }, { NULL, NULL } };
static struct FunctionField argsUTF8WriteValue[]        = { { "TotalChars",   ARG_LONG },   { "Value",   ARG_LONG   }, { "String", ARG_STRING }, { "StringSize", ARG_LONG }, { NULL, NULL } };

static struct Function JumpTableV1[] = {
   { IntToStr,              "IntToStr",              argsIntToStr },
   { StrShrink,             "StrShrink",             argsStrShrink },
   { StrExpand,             "StrExpand",             argsStrExpand },
   { StrInsert,             "StrInsert",             argsStrInsert },
   { StrSearch,             "StrSearch",             argsStrSearch },
   { StrUpper,              "StrUpper",              argsStrUpper },
   { StrLower,              "StrLower",              argsStrLower },
   { StrCapitalise,         "StrCapitalise",         argsStrCapitalise },
   { StrLineLength,         "StrLineLength",         argsStrLineLength },
   { StrNextLine,           "StrNextLine",           argsStrNextLine },
   { StrReplace,            "StrReplace",            argsStrReplace },
   { StrTranslate,          "StrTranslate",          argsStrTranslate },
   { StrCalculate,          "StrCalculate",          argsStrCalculate },
   { FloatToStr,            "FloatToStr",            argsFloatToStr },
   { StrToFloat,            "StrToFloat",            argsStrToFloat },
   { StrCopy,               "StrCopy",               argsStrCopy },
   { StrClone,              "StrClone",              argsStrClone },
   { StrLength,             "StrLength",             argsStrLength },
   { StrToInt,              "StrToInt",              argsStrToInt },
   { StrSort,               "StrSort",               argsStrSort },
   { StrBuildArray,         "StrBuildArray",         argsStrBuildArray },
   { UTF8CharacterOffset,   "UTF8CharacterOffset",   argsUTF8CharacterOffset },
   { UTF8Length,            "UTF8Length",            argsUTF8Length },
   { UTF8OffsetToCharacter, "UTF8OffsetToCharacter", argsUTF8OffsetToCharacter },
   { UTF8PrevLength,        "UTF8PrevLength",        argsUTF8PrevLength },
   { UTF8CharacterLength,   "UTF8CharacterLength",   argsUTF8CharacterLength },
   { UTF8ReadValue,         "UTF8ReadValue",         argsUTF8ReadValue },
   { UTF8WriteValue,        "UTF8WriteValue",        argsUTF8WriteValue },
   { StrFormat,             "StrFormat",             argsStrFormat },
   { StrFormatDate,         "StrFormatDate",         argsStrFormatDate },
   { StrReadColour,         "StrReadColour",         argsStrReadColour },
   { StrDatatype,           "StrDatatype",           argsStrDatatype },
   { CharCopy,              "CharCopy",              argsCharCopy },
   { NULL, NULL }
};

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, CMDOpen, NULL,
   JMP_DEFAULT, JumpTableV1, CPU_PC, MODVERSION_STRINGS, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1998-2004.  All rights reserved.",
   "January 2004",
   "Strings"
};

static struct KernelBase *KernelBase = 0;

static ERROR InsertString(STRING, STRING, LONG, LONG, LONG);

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;
   return(ERR_Okay);
}

/*****************************************************************************
** Command: Open()
*/

ERROR CMDOpen(OBJECTPTR Module)
{
   SetField(Module, FID_FunctionList, FT_POINTER, JumpTableV1);
   return(ERR_Okay);
}

/****************************************************************************/

inline static UBYTE UCase(UBYTE Case)
{
  if ((Case >= 'a') AND (Case <= 'z')) Case -= 0x20;
  return(Case);
}

inline static BYTE sortcompare(STRING Name1, STRING Name2)
{
   UBYTE char1, char2;

   while ((*Name1) AND (*Name2)) {
      char1 = *Name1;
      char2 = *Name2;
      if ((char1 >= 'A') AND (char1 <= 'Z')) char1 = char1 - 'A' + 'a';
      if ((char2 >= 'A') AND (char2 <= 'Z')) char2 = char2 - 'A' + 'a';

      if ((char1 > char2) OR ((Name1[1]) AND (!Name2[1]))) return(1); /* Name1 is greater */
      else if (char1 < char2) return(-1); /* Name1 is lesser */

      Name1++;
      Name2++;
   }

   return(0);
}

/*****************************************************************************

###FUNCTION###
Name:     CharCopy()
Short:    Copies the characters of one string to another.
Synopsis: LONG CharCopy(STRING Source, STRING Destination, LONG Length)

###DESCRIPTION###
<p>This function copies a string of characters to a destination.  It will
copy the exact number of characters as specified by Length, unless the
source String terminates before the total number of characters have been
copied.</p>

<p>This function will not null-terminate the destination.  This function
is not 'safe', in that it is capable of overflowing the destination buffer.  Use StrCopy()
if you want to copy a complete string with a limit on the destination buffer size.</p>

###INPUT###
Source:      Pointer to the string that you are copying from.
Destination: Pointer to the buffer that you are copying to.
Length:      The number of characters to copy.

###RESULT###
<p>Returns the total amount of characters that were copied.</p>

###SEE ALSO###
Functions: StrCopy, StrReplace, StrInsert

###END###

*****************************************************************************/

static LONG CharCopy(STRING String, STRING Dest, LONG Length)
{
   LONG i;

   if ((String) AND (Dest)) {
      for (i=0; (i < Length) AND (String[i]); i++) Dest[i] = String[i];
      return(i);
   }
   else return(0);
}

/*****************************************************************************

###FUNCTION###
Name:     FloatToStr()
Synopsis: LONG FloatToStr(DOUBLE Number, STRING String, LONG Size)
Short:    Converts a floating point number to a string.

###DESCRIPTION###
<p>This function converts 64-bit floating point numbers into strings.  It
supports both negative and positive numbers (in the case of negative numbers,
the resulting string will lead with a minus sign).</p>

<p>You are required to supply your own String buffer to this function and
define the Size of that buffer.  If the floating point number cannot fit in
the amount of space that you have provided, remaining digits will be
cut-off.</p>

###INPUT###
Number: The floating point number that you want to convert.
String: Points to a string that must be long enough to hold all the
        characters that the number will translate to.
Size:   The byte-size of the destination String buffer.

###RESULT###
<p>Returns the total amount of characters that were written to the string,
not including the null byte at the end.  Zero is returned if an error
occurred.</p>

###END###

*****************************************************************************/

static LONG FloatToStr(DOUBLE Float, STRING String, LONG StringSize)
{
   LONG chars;
   LARGE convert;

   if ((!String) OR (StringSize < 1)) return(0);

   chars = 0;
   convert = (LARGE)Float;
   if (convert < 0) convert = -convert;
   if ((Float < 0) AND (StringSize > 1)) String[chars++] = '-';

   chars += IntToStr(convert, String+chars, StringSize);

   Float = (Float - ((LARGE)Float)); /* Get rid of the whole part of the number */
   if (Float < 0) Float = -Float;

   if ((Float > 0) AND (chars < StringSize-2)) {
      String[chars++] = '.';
      if (StringSize+chars > chars+8) StringSize = chars+8;
      Float = Float * 10;
      while ((Float > 0.00001) AND (chars < StringSize-1)) {
         String[chars++] = ((LARGE)Float) + '0';
         Float = (Float - (LARGE)Float) * 10;
      }
   }

   String[chars] = 0;

   return(chars);
}

/*****************************************************************************

###FUNCTION###
Name:     IntToStr()
Synopsis: LONG IntToStr(LARGE Integer, STRING String, LONG StringSize)
Short:    Converts an integer to a string.

###DESCRIPTION###
<p>This function converts integers into strings.  It supports both negative
and positive numbers (in the case of negative numbers, the resulting string
will lead with a minus sign).</p>

<p>You are required to supply your own String buffer to this function and
define the Size of that buffer.  If the integer value cannot fit in the
amount of space that you have provided, remaining digits will be cut-off.</p>

###INPUT###
Integer: The value that you want to convert into a string.
String:  Points to a string that must be long enough to hold all the
         characters that the Integer will translate to.
Size:    The byte-size of the destination String buffer.

###RESULT###
<p>Returns the total amount of characters that were written to the string, not
including the null byte at the end.  Zero is returned if an error occurred.</p>

###END###

*****************************************************************************/

static LONG IntToStr(LARGE Integer, STRING String, LONG StringSize)
{
   LARGE count;
   LONG build, digits;

   if (!String) return(0);

   if (StringSize < 1) return(0);

   digits = 0;
   if (Integer < 0) {
      *String++ = '-';
      Integer   = -Integer;
      digits++;
   }

   if (Integer) {
      count = 1000000000;
      if (Integer < count) while (Integer < count) count = count/10;

      while ((count > 0) AND (StringSize > 1)) {
         build = 0;
         while (Integer >= count) {
            Integer -= count;
            build++;
         }
         *String++ = (BYTE)(build + '0');
         count = count/10;
         digits++;
         StringSize--;
      }
   }
   else if (StringSize > 1) {
      *String++ = '0';
      StringSize--;
      digits++;
   }

   if (StringSize > 0) *String = 0;
   return(digits);
}

/*****************************************************************************

###FUNCTION###
Name:     StrBuildArray()
Short:    Builds an array of strings from a sequential string list.
Synopsis: STRING * StrBuildArray(STRING List, LONG Size, LONG AmtEntries, LONG Flags)

###DESCRIPTION###
<p>This function is helpful for converting a buffer of sequential strings
into a more easily managed string array.  A &quot;sequentially arranged list
of strings&quot; is basically a number of strings arranged one after the
other in a single byte array.  It is similar in arrangement to a CSV file,
except the comma separators are replaced with null-terminators.</p>

<p>To convert such a string list into an array, you need to know the total
byte size of the list, as well as the total number of strings in the list.
If you don't know this information, you will need to alter your routine to
make provisions for it.</p>

<p>Once you call this function, it will allocate a memory block to contain
the array and string information.  The list will then be converted into the
array, which will be terminated with a NULL pointer at its end.  If you
have specified the SBF_SORT or SBF_NODUPLICATES Flags then the array will
be sorted into alphabetical order for your convenience.  The SBF_NODUPLICATES
flag also removes duplicated strings from the array.</p>

<p>Remember to free the allocated array when you are finished with it.</p>

###INPUT###
List:       Pointer to a list of sequentially arranged strings.
Size:       The total byte size of the List.
AmtEntries: The total number of strings specified in the List.
Flags:      Set to SBF_SORT to sort the list, or SBF_NODUPLICATES to sort the
            list and remove duplicated strings.

###RESULT###
<p>Returns an array of STRING pointers, or NULL if failure.</p>

###END###

*****************************************************************************/

static STRING * StrBuildArray(STRING List, LONG Size, LONG AmtEntries, LONG Flags)
{
   LONG i, j, pos;
   STRING *array, str;

   if ((List) AND (Size) AND (AmtEntries > 0)) {
      if (AllocMemory(Size + ((AmtEntries + 1) * sizeof(STRING)), MEM_DATA, (APTR *)&array, NULL) IS ERR_Okay) {
         str = (STRING)(array + AmtEntries + 1);
         for (i=0; i < Size; i++) str[i] = List[i];
         pos = 0;
         for (i=0; i < AmtEntries; i++) {
            array[i] = str+pos;
            while (str[pos]) pos++;
            pos++;
            if (pos > Size) {
               DPrintF("@StrBuildArray()","The string buffer exceeds its specified length of %d bytes.", Size);
               break;
            }
         }
         array[i] = NULL;

         if (Flags & SBF_NODUPLICATES) {
            StrSort(array, NULL);
            for (i=1; array[i]; i++) {
               if (StrCompare(array[i-1], array[i], NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  for (j=i; array[j]; j++) array[j] = array[j+1];
                  i--;
                  AmtEntries--;
               }
            }
            array[AmtEntries] = NULL;
         }
         else if (Flags & SBF_SORT) StrSort(array, NULL);

         return(array);
      }
      else return(NULL);
   }
   else return(NULL);
}

/*****************************************************************************

###FUNCTION###
Name:     StrCalculate()
Synopsis: ERROR StrCalculate(STRING String, DOUBLE *Result, STRING Buffer, LONG BufferSize)
Short:    Calculates the total of any string containing calculations.

###DESCRIPTION###
<p>This function can be used for performing simple calculations on numeric values
and strings.  It can return a result in either a numeric format or in a string
buffer if the calculation involves non-numeric characters.  Here are some
examples of valid strings:</p>

<pre>
   "100/50+(12*14)"
   "0.05 * 100 + '%'"
</pre>

<p>Currently acceptable operators are plus, minus, divide and multiply.  String
references must be enclosed in single quotes or will be ignored.  Brackets may
be used to organise the order of operations during calculation.</p>

<p>Special operators include:</p>

<pre>
  p    This character immediately followed with an integer allows you to change
       the floating-point precision of output values.
  f    The same as the 'p' operator except the precision is always guaranteed
       to be fixed at that value through the use of trailing zeros (so a fixed
       precision of two used to print the number '7' will give a result of
       '7.00'.
</pre>

<p>The StrCalculate() function uses 64-bit floating point operations for
maximum accuracy.</p>

###INPUT###
String: A string containing the calculation text.
Result: Refers to a DOUBLE variable that will store the calculated number, if you are only interested in calculating a final figure (optional).
Buffer: A result buffer is required if you want to interpret the resulting calculation as a string (recommended).
BufferSize: The size of the Buffer, in bytes.

###RESULT###
ERR_Okay:    Success - the Result contains the calculated floating point number.
ERR_BadData: The string that was supplied contains invalid characters.
ERR_Args:    Invalid arguments were specified.

###END###

*****************************************************************************/

#define SIGN_PLUS     1
#define SIGN_MINUS    2
#define SIGN_MULTIPLY 3
#define SIGN_DIVIDE   4

static WORD WriteCalculation(STRING Buffer, LONG BufferSize, DOUBLE Value, WORD Precision)
{
   LONG index;
   WORD px;
   LARGE convert;

   index = 0;
   convert = Value;
   if (convert < 0) convert = -convert;

   /*** Sign the value if it is less than 0 ***/

   if ((Value < 0) AND (index < BufferSize - 1)) Buffer[index++] = '-';

   index += IntToStr(convert, Buffer+index, BufferSize);

   Value = (Value - ((LARGE)Value)); /* Get rid of the whole part of the number */
   if (Value < 0) Value = -Value;

   if ((index < BufferSize-1) AND ((Value > 0) OR (Precision < 0))) {
      Buffer[index++] = '.';
      Value = Value * 10;
      px = Precision;
      if (px < 0) px = -px;
      while ((Value > 0.00001) AND (index < BufferSize-1) AND (px > 0)) {
         Buffer[index++] = ((LARGE)Value) + '0';
         Value = (Value - (LARGE)Value) * 10;
         px--;
      }

      if (Precision < 0) {
         while (px > 0) { Buffer[index++] = '0'; px--; }
      }
   }

   return(index);
}

static ERROR StrCalculate(STRING String, DOUBLE *Result, STRING Buffer, LONG BufferSize)
{
   UBYTE buffer[180], calc[20];
   DOUBLE total, calcfloat, overall;
   UBYTE sign, number;
   STRING alloc, newstring;
   LONG i, j, index, bracketpos;
   WORD precision;

   if ((!String) OR ((!Result) AND (!Buffer))) {
      DPrintF("@StrCalculate()","Missing arguments.");
      return(ERR_Args);
   }

   if (Result) *Result = NULL;

   if (Buffer) {
      if (BufferSize < 1) return(ERR_BufferOverflow);
      Buffer[0] = 0;
   }

   if (Buffer IS String) {
      DPrintF("@StrCalculate()","Input and output arguments point to the same address.");
      return(ERR_Args);
   }

   precision = 9;

   /*** Search for brackets and translate them first ***/

   alloc = NULL;
   while (1) {

      /*** Find the last bracketed reference ***/

      bracketpos = 0;
      for (i=0; String[i]; i++) {
         if (String[i] IS '\'') {
            /*** Skip anything that is in quotes ***/
            i++;
            while (String[i]) {
               if (String[i] IS '\\') {
                  i++; /* Skip backslashes the immediate character afterwards */
                  if (!String[i]) break;
               }
               else if (String[i] IS '\'') break;
               i++;
            }
            if (String[i] IS '\'') i++;
            continue;
         }
         if (String[i] IS '(') bracketpos = i;
      }

      /*** If we found a bracket, translate its contents ***/

      if (bracketpos > 0) {
         buffer[0] = ' ';
         j = 1;
         for (i=bracketpos+1; (String[i] != 0) AND (String[i] != ')'); i++) {
            buffer[j++] = String[i];
            if (j > sizeof(buffer)-3) break;
         }
         buffer[0] = '(';
         buffer[j++] = ')';
         buffer[j] = 0;
         StrCalculate(buffer+1, &calcfloat, NULL, NULL);
         FloatToStr(calcfloat, calc, sizeof(calc));

         if (StrReplace(String, buffer, calc, &newstring, TRUE) IS ERR_Okay) {
            if (alloc) FreeMemory(alloc);
            alloc = String = newstring;
         }
         else break;
      }
      else break;
   }

   /*** Perform the calculation ***/

   total = 0;
   overall = 0;
   index = 0;
   sign  = SIGN_PLUS;
   number = FALSE;
   while (*String) {
      if (*String <= 0x20); /* Do nothing with whitespace */
      else if (*String IS '\'') {
         if (Buffer) {
            if (number) {
               /*** Write the current floating point number to the buffer before we deal with the next calculation ***/

               index += WriteCalculation(Buffer+index, BufferSize - index, total, precision);

               /*** Reset the number ***/

               overall += total;
               total = 0;
               number = FALSE;
            }

            String++;
            while (index < BufferSize-1) {
               if (*String IS '\\') {
                  String++; /* Skip the \ character and continue so that we can copy the character immediately after it */
               }
               else if (*String IS '\'') break;

               Buffer[index++] = *String;
               String++;
            }
         }
         else {
            /*** Skip string content if there is no string buffer ***/
            String++;
            while (*String != '\'') String++;
         }
      }
      else if (*String IS 'f') {
         /*** Fixed floating point precision adjustment ***/
         String++;
         precision = -StrToInt(String);
         while ((*String >= '0') AND (*String <= '9')) String++;
         continue;
      }
      else if (*String IS 'p') {
         /*** Floating point precision adjustment ***/
         String++;
         precision = StrToInt(String);
         while ((*String >= '0') AND (*String <= '9')) String++;
         continue;
      }
      else if ((*String >= '0') AND (*String <= '9')) {
         number = TRUE;
         if (sign IS SIGN_MINUS)         total = total - StrToFloat(String);
         else if (sign IS SIGN_MULTIPLY) total = total * StrToFloat(String);
         else if (sign IS SIGN_DIVIDE)   total = total / StrToFloat(String);
         else total += StrToFloat(String);
         while (((*String >= '0') AND (*String <= '9')) OR (*String IS '.')) String++;

         sign = SIGN_PLUS; /* The mathematical sign is reset whenever a number is encountered */
         continue;
      }
      else if (*String IS '-') {
         if (sign IS SIGN_MINUS) sign = SIGN_PLUS; /* Handle double-negatives */
         else sign = SIGN_MINUS;
      }
      else if (*String IS '+') sign = SIGN_PLUS;
      else if (*String IS '*') sign = SIGN_MULTIPLY;
      else if (*String IS '/') sign = SIGN_DIVIDE;

      for (++String; (*String & 0xc0) IS 0x80; String++);
   }

   if (Buffer) {
      if (number) index += WriteCalculation(Buffer+index, BufferSize - index, total, precision);
      Buffer[index] = 0;
   }

   if (alloc) FreeMemory(alloc);
   if (Result) *Result = overall + total;
   return(ERR_Okay);
}

/*****************************************************************************

###FUNCTION###
Name:     StrCapitalise()
Short:    Capitalises a string.
Synopsis: void StrCapitalise(STRING String)

###DESCRIPTION###
<p>This function will will capitalise a string so that every word starts
with a capital letter.  All letters following the first capital of each word
will be driven to lower-case characters.  Numbers and other non-alphabetic
characters will not be affected by this function.  Here is an example:</p>

<pre>
   "every WOrd starts WITH a 2apital" = "Every Word Starts With A 2apital"
</pre>

###INPUT###
String: Points to the string that you want to capitalise.

###END###

*****************************************************************************/

static void StrCapitalise(STRING String)
{
  WORD space;

  space = TRUE;
  while (*String) {
     if (space) {
        if ((*String >= 'a') AND (*String <= 'z')) *String -= 0x20;
        else if (*String != ' ') space = FALSE;
     }
     else {
        if ((*String >= 'A') AND (*String <= 'Z')) *String += 0x20;
        else if (*String IS ' ') space = TRUE;
     }
     for (++String; (*String & 0xc0) IS 0x80; String++);
  }
}

/*****************************************************************************

###FUNCTION###
Name:     StrClone()
Short:    Clones string data.
Synopsis: STRING StrClone(STRING String)

###DESCRIPTION###
<p>This function allows you to create an exact duplicate of a string.  It
analyses the length of the supplied String, allocates a private memory block
for it and then copies the string characters into the new string buffer.</p>

<p>You are expected to free the resulting memory block when you are finished
with it.</p>

###INPUT###
String: Pointer to the string that you want to clone.

###RESULT###
<p>Returns an exact duplicate of the String.  If this function fails, NULL is
returned.</p>

###END###

*****************************************************************************/

static STRING StrClone(STRING String)
{
   LONG i;
   STRING newstr;

   if (!String) return(NULL);

   for (i=0; String[i]; i++);

   if ((AllocMemory(i+1, MEM_STRING, (void **)&newstr, NULL)) IS ERR_Okay) {
      for (i=0; String[i]; i++) newstr[i] = String[i];
      newstr[i] = 0;
      return(newstr);
   }
   else return(NULL);
}

/*****************************************************************************

###FUNCTION###
Name:     StrCopy()
Short:    Copies the characters of one string to another.
Synopsis: LONG StrCopy(STRING Source, STRING Destination, LONG Length)

###DESCRIPTION###
<p>This function copies part of one string over to another.  If the Length is
set to zero then this function will copy the entire Source string over to
the Destination.  Note that if this function encounters the end of the Source
string (i.e. the null byte) while copying, then it will stop automatically
to prevent copying of junk characters.</p>

<p>Please note that the Destination string will <i>always</i> be null-terminated by
this function regardless of whether you set the Length or not.  For example, if
you were to copy &quot;123&quot; into the middle of
string &quot;ABCDEFGHI&quot; then the result would be &quot;ABC123&quot;.
The "GHI" part of the string would be lost.  In situations such as this,
functions such as <a href="#CharCopy()">CharCopy()</a>, <a href="#StrReplace()">StrReplace()</a> or
<a href="#StrInsert()">StrInsert()</a> should be used instead.</p>

###INPUT###
Source:      Pointer to the string that you are copying from.
Destination: Pointer to the buffer that you are copying to.
Length:      The maximum number of characters to copy, including the NULL byte.  The Length typically indicates the buffer size of the Destination pointer.  Setting the Length to COPY_ALL will copy all characters from the Source.

###RESULT###
<p>Returns the total amount of characters that were copied, not including the
null byte at the end.</p>

###SEE ALSO###
Functions: CharCopy, StrReplace, StrInsert

###END###

*****************************************************************************/

static LONG StrCopy(STRING String, STRING Dest, LONG Length)
{
   LONG i;

   if (Length < 0) return(0);

   i = 0;
   if ((String) AND (Dest)) {
      if (!Length) DPrintF("@StrCopy()","Warning - zero length given for string \"%s\".", String);

      if (!Length) Length = 0x7fffffff; /* Largest number */

      while ((i < Length) AND (*String)) {
         Dest[i++] = *String++;
      }

      if ((*String) AND (i >= Length)) {
         DPrintF("@StrCopy()","Overflow: %d/%d \"%s\"", i, Length, Dest);
         Dest[i-1] = 0; /* If we ran out of buffer space, we have to terminate from one character back */
      }
      else Dest[i] = 0;
   }

   return(i);
}

/*****************************************************************************

###FUNCTION###
Name:     StrDatatype()
Short:    Determines the data type of a string.
Synopsis: LONG StrDatatype(STRING String)

###DESCRIPTION###
<p>This function analyses a string and returns its data type.  Valid return
values are STR_FLOAT for floating point numbers, STR_NUMBER for whole
numbers and STR_STRING for any other string type.  In order for the string
to be recognised as one of the number types, it must be limited to numbers
and qualification characters, such as a decimal point or negative sign.</p>

###INPUT###
String: The string that you want to analyse.

###RESULT###
<p>Returns STR_FLOAT, STR_NUMBER or STR_STRING.</p>

###END###

*****************************************************************************/

static LONG StrDatatype(STRING String)
{
   LONG i, number, Float;

   number = TRUE;
   Float  = FALSE;
   for (i=0; (String[i]) AND (number IS TRUE); i++) {
      if (String[i] <= 0x20) continue; /* Skip whitespace */
      if (((String[i] < '0') OR (String[i] > '9')) AND
          (String[i] != '.') AND (String[i] != '-')) number = FALSE;
      if (String[i] IS '.') Float = TRUE;
   }

   if ((Float) AND (number)) return(STR_FLOAT);
   else if (number) return(STR_NUMBER);
   else return(STR_STRING);
}

/*****************************************************************************

###FUNCTION###
Name:     StrExpand()
Synopsis: LONG StrExpand(STRING String, LONG Position, LONG TotalChars)
Short:    Expands the size of a string by inserting spaces.

###DESCRIPTION###
<p>This function will expand a string by inserting spaces into a specified
position.  The String that you are expanding must be placed in a memory area
large enough to accept the increased size, or you will almost certainly
corrupt other memory areas.</p>

###INPUT###
String:     The string that you want to expand.
Position:   The byte position that you want to start expanding from.  Give consideration to UTF-8 formatting.
TotalChars: The total number of spaces that you want to insert.

###RESULT###
<p>Returns the new length of the expanded string, not including the
null byte.</p>

###END###

*****************************************************************************/

static LONG StrExpand(STRING String, LONG Pos, LONG AmtChars)
{
   LONG Len, i;

   if ((String) AND (AmtChars)) {
      if ((Len = StrLength(String)) > 0) {
         if (Pos < 0) Pos = 0;
         if (Pos > Len) Pos = Len;

         /*** Shift the characters at Pos to the right ***/

         String[Len + AmtChars] = 0;        /* Set new termination */
         for (i = Len-1; i > Pos-1; i--) {  /* Shift the characters across */
            String[i + AmtChars] = String[i];
         }

         for (i = Pos; i < (Pos + AmtChars); i++) String[i] = ' ';

         return(Len + AmtChars);
      }
   }
   else DPrintF("!StrExpand:","Bad arguments.");

   return(0);
}

/*****************************************************************************

###FUNCTION###
Name:     StrFormat()
Synopsis: LONG StrFormat(STRING Buffer, LONG BufferSize, STRING Format, Args...);
Short:    Formats a string using printf() style arguments.

###DESCRIPTION###
<p>Undocumented.</p>

###INPUT###
Buffer: Reference to a destination buffer for the formatted string.
Length: The length of the destination Buffer, in bytes.
Format: The string format to use for processing.
Args:   A tag-list of arguments that match the parameters in the Format string.

###RESULT###
<p>Returns the total number of characters written to the buffer (not including
the NULL terminator).  If the total number of characters is greater or equal
to the BufferSize minus 1, then you should assume that a buffer overflow
occurred.</p>

###END###

*****************************************************************************/

static LONG StrFormat(STRING Buffer, LONG BufferSize, STRING String, STRING Args, ...)
{
   LONG *array;

   if (!String) return(0);

   if (BufferSize <= 0) BufferSize = 0x7fff;

   array = (LONG *)&Args;
   #ifdef _WIN32
      _snprintf(Buffer, BufferSize, String, array[0], array[1], array[2], array[3], array[4], array[5], array[6], array[7], array[8], array[9], array[10], array[11], array[12], array[13], array[14], array[15], array[16], array[17], array[18]);
   #else
      snprintf(Buffer, BufferSize, String, array[0], array[1], array[2], array[3], array[4], array[5], array[6], array[7], array[8], array[9], array[10], array[11], array[12], array[13], array[14], array[15], array[16], array[17], array[18]);
   #endif

   return(StrLength(Buffer));
}

/*****************************************************************************

###FUNCTION###
Name:     StrFormatDate
Synopsis: ERROR StrFormatDate(STRING Buffer, LONG Length, STRING Format, *Time)
Short:    Formats the content of a Time object into a date string.

###DESCRIPTION###
<p>The StrFormatDate() function is used to format dates into a human readable
string format.  The Format argument defines how the time information is
printed to the display.  Time formatting follows accepted conventions, as
illustrated in the following format table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">String</td><td class="tableheader" bgcolor="#eeeeee">Example</td></tr>
<tr><td width="10%">h</td><td>24-Hour: 9</td></tr>
<tr><td width="10%">hh</td><td>24-Hour: 09</td></tr>
<tr><td width="10%">H</td><td>12-Hour: 9</td></tr>
<tr><td width="10%">HH</td><td>12-Hour: 09</td></tr>
<tr><td width="10%">n</td><td>Minute: 7</td></tr>
<tr><td width="10%">nn</td><td>Minute: 07</td></tr>
<tr><td width="10%">s</td><td>Second: 42</td></tr>
<tr><td width="10%">ss</td><td>Second: 42</td></tr>
<tr><td width="10%">d</td><td>Day: 3</td></tr>
<tr><td width="10%">dd</td><td>Day: 03</td></tr>
<tr><td width="10%">D</td><td>Day of week: T</td></tr>
<tr><td width="10%">ddd</td><td>Day of week: sat</td></tr>
<tr><td width="10%">dddd</td><td>Day of week: saturday</td></tr>
<tr><td width="10%">m</td><td>Month: 11</td></tr>
<tr><td width="10%">mm</td><td>Month: 11</td></tr>
<tr><td width="10%">mmm</td><td>Month: nov</td></tr>
<tr><td width="10%">mmmm</td><td>Month: november</td></tr>
<tr><td width="10%">yy</td><td>Year: 03</td></tr>
<tr><td width="10%">yyyy</td><td>Year: 2003</td></tr>
</table>

<p>By default, strings are printed in lower-case.  For example, "ddd" prints
"sat".  If you want to capitalise the characters, use capitals in the string
formatting instead of the default lower case.  For example "Ddd" prints "Sat"
and "DDD" prints "SAT".</p>

<p>Where the interpreter encounters unrecognised characters, it will print
them out unaltered.  For example "hh:nn:ss" might print "05:34:19" with the
colons intact.  If you need to include characters such as 'h' or 'd' and
want to prevent them from being translated, preceed them with the backslash
character.  For instance to print "0800 hours", we would use the format
"hh00 \hours".</p>

<p>The Time structure that you provide needs to contain values in the
Year, Month, Day, Hour, Minute and Second fields.  If you know that you
do not plan on printing information on any particular date value, you can
optionally choose not to set that field.</p>

###INPUT###
Buffer: Pointer to a buffer large enough to receive the formatted data.
Length: The length of the Buffer, in bytes.
Format: String specifying the required date format.
Time:   Pointer to a Time structure that contains the date to be formatted.

###RESULT###
ERR_Okay: The date was formatted successfully.
ERR_Args: Invalid arguments were specified.
ERR_BufferOverflow: The buffer was not large enough to hold the formatted string (the result will be correct, but truncated).

###END###

*****************************************************************************/

static STRING days[7] = { "sunday", "monday", "tuesday", "wednesday", "thursday", "friday", "saturday" };
static STRING DAYS[7] = { "SUNDAY", "MONDAY", "TUESDAY", "WEDNESDAY", "THURSDAY", "FRIDAY", "SATURDAY" };

static STRING months[13] = { "", "january", "february", "march", "april", "may", "june", "july", "august", "september", "october", "november", "december" };
static STRING MONTHS[13] = { "", "JANUARY", "FEBRUARY", "MARCH", "APRIL", "MAY", "JUNE", "JULY", "AUGUST", "SEPTEMBER", "OCTOBER", "NOVEMBER", "DECEMBER" };

static ERROR StrFormatDate(STRING Buffer, LONG BufferSize, STRING Format, struct Time *Time)
{
   STRING str;
   WORD tmp, pos, i;

   if ((!Buffer) OR (BufferSize < 1) OR (!Format) OR (!Time)) {
      return(ERR_Args);
   }

   str = Buffer;
   pos = 0;
   for (i=0; (Format[i]) AND (pos < BufferSize-1); i++) {
      if (Format[i] IS '\\') {
         if (Format[i+1]) {
            str[pos++] = Format[i+1];
            i++;
         }
         else str[pos++] = '\\';
      }
      else if (Format[i] IS 'h') {
         /*** 24 hour clock ****/

         if (Format[i+1] IS 'h') {
            i++;
            if (Time->Hour < 10) str[pos++] = '0';
         }
         pos += IntToStr(Time->Hour, str + pos, BufferSize-pos-1);
      }
      else if (Format[i] IS 'H') {
         /*** 12 hour clock ***/

         if (Format[i+1] IS 'H') {
            i++;
            if (Time->Hour < 10) str[pos++] = '0';
         }
         if (Time->Hour > 12) pos += IntToStr(Time->Hour - 12, str + pos, BufferSize-pos-1);
         else pos += IntToStr(Time->Hour, str + pos, BufferSize-pos-1);
      }
      else if (Format[i] IS 'n') {
         /*** Minutes ***/

         if (Format[i+1] IS 'n') {
            i++;
            if (Time->Minute < 10) str[pos++] = '0';
         }
         pos += IntToStr(Time->Minute, str + pos, BufferSize-pos-1);
      }
      else if (Format[i] IS 's') {
         /*** Seconds ***/

         if (Format[i+1] IS 's') {
            i++;
            if (Time->Second < 10) str[pos++] = '0';
         }
         pos += IntToStr(Time->Second, str + pos, BufferSize-pos-1);
      }
      else if (Format[i] IS 'y') {
         /*** Year ***/

         if ((Format[i+1] IS 'y') AND (Format[i+2] != 'y')) {
            tmp = Time->Year % 100;
            if (tmp < 10) str[pos++] = '0';
            pos += IntToStr(tmp, str + pos, BufferSize-pos-1);
         }
         else pos += IntToStr(Time->Year, str + pos, BufferSize-pos-1);

         while (Format[i+1] IS 'y') i++;
      }
      else if ((Format[i] IS 'd') OR (Format[i] IS 'D')) {
         /* Day: d    = Single digit
         **      D    = Day of Week character
         **      dd   = Double digit
         **      ddd  = Short name
         **      dddd = Long name
         **      Ddd  = Short name caps
         **      Dddd = Long name caps
         */

         if ((Format[i+1] != 'd') AND (Format[i+2] != 'D')) {
            if (Format[i] IS 'D') str[pos++] = DAYS[Time->DayOfWeek][0];
            else pos += IntToStr(Time->Day, str + pos, BufferSize-pos-1);
         }
         else if ((Format[i+2] != 'd') AND (Format[i+2] != 'D')) {
            i++;
            if (Time->Day < 10) str[pos++] = '0';
            pos += IntToStr(Time->Day, str + pos, BufferSize-pos-1);
         }
         else if ((Format[i+3] != 'd') AND (Format[i+3] != 'D')) {
            if (Format[i] IS 'D') {
               if (Format[i+1] IS 'D') {
                  pos += CharCopy(DAYS[Time->DayOfWeek], str + pos, 3);
               }
               else {
                  str[pos++] = DAYS[Time->DayOfWeek][0];
                  pos += CharCopy(days[Time->DayOfWeek]+1, str + pos, 2);
               }
            }
            else pos += CharCopy(days[Time->DayOfWeek], str + pos, 3);
            i += 2;
         }
         else if ((Format[i+4] != 'd') AND (Format[i+4] != 'D')) {
            if (Format[i] IS 'D') {
               if (Format[i+1] IS 'D') {
                  pos += StrCopy(DAYS[Time->DayOfWeek], str + pos, BufferSize-pos);
               }
               else {
                  str[pos++] = DAYS[Time->DayOfWeek][0];
                  pos += StrCopy(days[Time->DayOfWeek]+1, str + pos, BufferSize-pos);
               }
            }
            else pos += StrCopy(days[Time->DayOfWeek], str + pos, BufferSize-pos);

            i += 3;
         }
      }
      else if ((Format[i] IS 'm') OR (Format[i] IS 'M')) {
         /* Month: m    = Single digit
         **        mm   = Double digit
         **        mmm  = Short name
         **        mmmm = Long name
         **        Mmm  = Short name caps
         **        Mmmm = Long name caps
         */

         if ((Format[i+1] != 'm') AND (Format[i+2] != 'M')) {
            pos += IntToStr(Time->Month, str + pos, BufferSize-pos-1);
         }
         else if ((Format[i+2] != 'm') AND (Format[i+2] != 'M')) {
            i++;
            if (Time->Month < 10) str[pos++] = '0';
            pos += IntToStr(Time->Month, str + pos, BufferSize-pos-1);
         }
         else if ((Format[i+3] != 'm') AND (Format[i+3] != 'M')) {
            if (Format[i] IS 'M') {
               if (Format[i+1] IS 'M') {
                  pos += CharCopy(MONTHS[Time->Month], str + pos, 3);
               }
               else {
                  str[pos++] = MONTHS[Time->Month][0];
                  pos += CharCopy(months[Time->Month]+1, str + pos, 2);
               }
            }
            else pos += CharCopy(months[Time->Month], str + pos, 3);
            i += 2;
         }
         else if ((Format[i+4] != 'm') AND (Format[i+4] != 'M')) {
            if (Format[i] IS 'M') {
               if (Format[i+1] IS 'M') {
                  pos += StrCopy(MONTHS[Time->Month], str + pos, BufferSize-pos);
               }
               else {
                  str[pos++] = MONTHS[Time->Month][0];
                  pos += StrCopy(months[Time->Month]+1, str + pos, BufferSize-pos);
               }
            }
            else pos += StrCopy(months[Time->Month], str + pos, BufferSize-pos);

            i += 3;
         }
      }
      else str[pos++] = Format[i];
   }
   str[pos] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FUNCTION###
Name:     StrInsert()
Synopsis: ERROR StrInsert(STRING Insert, STRING Buffer, LONG Position, LONG ReplaceChars, LONG BufferSize)
Short:    Inserts a string into a buffer, with optional replace.

###DESCRIPTION###
<p>This function is used to insert a series of characters into a
Buffer.  The position of the insert is determined by the byte offset
declared in the Position field.  You can opt to replace a specific number
of characters in the destination by setting the ReplaceChars argument to a
value greater than zero.  To prevent buffer over-runs, you must declare
the size of the Buffer in the BufferSize argument.</p>

###INPUT###
Insert:       The string characters that you want to insert.
Buffer:       The string that you want to insert the characters into.
Position:     The byte position at which the insertion will start.  Give consideration to UTF-8 formatting.
ReplaceCount: The number of characters to replace (set to zero for a normal insert).
BufferSize:   The byte size of the Buffer.

###RESULT###
ERR_Okay: Insertion successful.
ERR_Args: Invalid arguments were specified.
ERR_BufferOverflow: There is not enough space in the destination buffer for the insert.

###END###

*****************************************************************************/

static ERROR StrInsert(STRING Insert, STRING Buffer, LONG Pos, LONG ReplaceChars, LONG BufferSize)
{
   LONG insertlen, i, strlen, j;

   if (!Insert) Insert = "";

   for (insertlen=0; Insert[insertlen]; insertlen++);

   /*** String insertion ***/

   if (insertlen < ReplaceChars) {
      i = Pos + StrCopy(Insert, Buffer+Pos, COPY_ALL);
      i = Pos + ReplaceChars;
      Pos += insertlen;
      while (Buffer[i]) Buffer[Pos++] = Buffer[i++];
      Buffer[Pos] = NULL;
   }
   else if (insertlen IS ReplaceChars) {
      while (*Insert) Buffer[Pos++] = *Insert++;
   }
   else {
      /*** Check if an overflow will occur ***/

      for (strlen=0; Buffer[strlen]; strlen++);
      if ((BufferSize - 1) < (strlen + (ReplaceChars - insertlen))) {
         DPrintF("@StrInsert()","Buffer overflow: \"%.60s\"", Buffer);
         return(ERR_BufferOverflow);
      }

      /*** Expand the string ***/
      i = strlen + (insertlen - ReplaceChars) + 1;
      strlen += 1;
      j = strlen-Pos-ReplaceChars+1;
      while (j > 0) {
         Buffer[i--] = Buffer[strlen--];
         j--;
      }

      /*** Copy the insert string into the position ***/
      for (i=0; Insert[i]; i++, Pos++) Buffer[Pos] = Insert[i];
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FUNCTION###
Name:     StrLength()
Short:    Calculates the length of a string.
Synopsis: LONG StrLength(STRING String)

###DESCRIPTION###
<p>This function will calculate the length of a String, not including the
null byte.</p>

###INPUT###
String: Pointer to the string that you want to examine.

###RESULT###
<p>Returns the length of the string.</p>

###SEE ALSO###
Function: StrLineLength()

###END###

*****************************************************************************/

static LONG StrLength(STRING String)
{
   LONG i;
   if (String) {
      for (i = 0; String[i]; i++);
      return(i);
   }
   else return(0);
}

/*****************************************************************************

###FUNCTION###
Name:     StrLineLength()
Synopsis: LONG StrLineLength(STRING String)
Short:    Determines the line-length of a string.

###DESCRIPTION###
<p>This is a simple function that calculates the length of a String up to the
first carriage return, line-break or null byte.  This function is commonly
used on large text files where the String points to a position inside a
large character buffer.</p>

###INPUT###
String: Pointer to an address inside a character buffer.

###RESULT###
<p>Returns the length of the first line in the string, not including the
return code.</p>

###SEE ALSO###
Function: StrLength(), StrNextLine()

###END###

*****************************************************************************/

static LONG StrLineLength(STRING String)
{
   LONG i;

   if (String) {
      i = 0;
      while ((String[i] != 0) AND (String[i] != 10) AND (String[i] != 13)) i++;
      return(i);
   }
   else return(0);
}

/*****************************************************************************

###FUNCTION###
Name:     StrLower()
Short:    Changes a string so that all alpha characters are in lower-case.
Synopsis: void StrLower(STRING String)

###DESCRIPTION###
<p>This function will alter a string so that all upper case characters are
changed to lower-case.  Non upper-case characters are unaffected by this
function.  Here is an example:</p>

<pre>
   "HeLLo world" = "hello world"
</pre>

###INPUT###
String: Pointer to the string that you want to change to lower-case.

###END###

*****************************************************************************/

static void StrLower(STRING String)
{
   if (!String) return;

   while (*String) {
      if ((*String >= 'A') AND (*String <= 'Z')) *String += 0x20;
      String++;
   }
}

/*****************************************************************************

###FUNCTION###
Name:     StrNextLine()
Synopsis: STRING StrNextLine(STRING Buffer);
Short:    Returns a pointer to the next line in a string buffer.

###DESCRIPTION###
<p>This function scans a character buffer for a carriage return or line feed
and returns a pointer to the following line.  If no return code is found,
NULL is returned.</p>

<p>This function is commonly used for scanning text files on a line by line
basis.</p>

###INPUT###
Buffer: Pointer to an address inside a character buffer.

###RESULT###
<p>Returns a pointer to the string following the next line, or NULL if no
further lines are present.</p>

###END###

*****************************************************************************/

static STRING StrNextLine(STRING String)
{
   if (!String) return(NULL);

   while ((*String) AND (*String != '\n') AND (*String != '\r')) String++;
   while (*String IS '\r') String++;
   if (*String IS '\n') String++;
   while (*String IS '\r') String++;
   if (*String) return(String);
   else return(NULL);
}
/*****************************************************************************

###FUNCTION###
Name:     StrReadColour()
Short:    Converts a colour string into its equivalent RGB values.
Synopsis: ERROR ReadColour(STRING Colour, LONG *Red, LONG *Green, LONG *Blue, LONG *Alpha)

###DESCRIPTION###
<p>Use the StrReadColour() function to convert a colour from its string format
to equivalent red, green and blue values.  The colour that is referenced
must be in hexadecimal or separated-decimal format.  For example a
pure red colour may be expressed as a string of "#ff0000" or "255,0,0".</p>

###INPUT###
Colour: Pointer to a string containing the colour.
Red:    Pointer to a 32-bit value that will hold the red component.
Green:  Pointer to a 32-bit value that will hold the green component.
Blue:   Pointer to a 32-bit value that will hold the blue component.
Alpha:  Pointer to a 32-bit value that will hold the alpha component.

###RESULT###
ERR_Okay: The colour conversion was successful.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static WORD ReadByte(STRING Str)
{
   WORD Value;
   if (!*Str) return(0);
   if ((*Str >= '0') AND (*Str <= '9')) Value = (*Str - '0') * 16;
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value = ((*Str - 'A')+10) * 16;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value = ((*Str - 'a')+10) * 16;
   else Value = 0;
   Str++;
   if ((*Str >= '0') AND (*Str <= '9')) Value += (*Str - '0');
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value += (*Str - 'A') + 10;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value += (*Str - 'a') + 10;
   return(Value);
}

static ERROR StrReadColour(STRING Colour, LONG *Red, LONG *Green, LONG *Blue, LONG *Alpha)
{
   if ((!Colour) OR (!Red) OR (!Green) OR (!Blue)) return(ERR_Args);

   if (*Colour IS '#') {
      Colour++;
      *Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Blue  = ReadByte(Colour);
      if (Alpha) {
         if (*Colour) Colour++; if (*Colour) Colour++;
         if (*Colour) *Alpha = ReadByte(Colour);
         else *Alpha = 255;
      }
   }
   else {
      *Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Blue = StrToInt(Colour);
      if (Alpha) {
         while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
         while ((*Colour) AND (*Colour <= 0x20)) Colour++;
         if (*Colour) *Alpha = StrToInt(Colour);
         else *Alpha = 255;
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FUNCTION###
Name:     StrReplace()
Short:    Replaces all occurances of a keyword or phrase within a given string.
Synopsis: ERROR StrReplace(STRING Source, STRING Keyword, STRING Replacement, STRING *Result, LONG CaseSensitive)

###DESCRIPTION###
<p>This function will search a string and replace all occurances of a Keyword
or phrase with the supplied Replacement string.  If the Keyword is found, a
new string will be returned which has all matches replaced with the
Replacement string.</p>

<p>If the Keyword is not found, an error code of ERR_Search will be
returned.  If you want to know whether the Keyword actually exists before
performing a replacement, consider calling the
<a href="#StrSearch()">StrSearch()</a> function first.</p>

<p>The new string that is created will be stored in the Result variable as
a standard memory block.  You are expected to free the memory block once you
are finished with it.</p>

###INPUT###
Source:        Points to the source string that contains occurances of the
               Keyword that you are searching for.
Keyword:       Identifies the keyword or phrase that you want to replace.
Replacement:   Identifies the string that will replace all occurances of the Keyword.
Result:        Must refer to a STRING variable that will store the resulting memory block.
CaseSensitive: Set to TRUE if the keyword search should be case-sensitive.

###RESULT###
ERR_Okay:        A replacement string was found and replaced.
ERR_Args:        Invalid arguments were specified.
ERR_Search:      The Keyword could not be found in the Source string.
ERR_AllocMemory: Memory for the resulting string could not be allocated.

###END###

*****************************************************************************/

static ERROR StrReplace(STRING Source, STRING Keyword, STRING Replacement, STRING *Result,
                LONG CaseSensitive)
{
   LONG pos, i, j, KeywordLength, ReplaceLength, NewSize, offset;
   WORD Alloc;
   STRING NewString, Original;

   *Result = NULL;

   if ((!Source) OR (!Keyword) OR (!Result)) {
      return(SendError(ERH_StrReplace, ERR_Args));
   }

   if (!Replacement) Replacement = "";

   /* If the Keyword and the replacement are identical,
   ** there is no need to do any replacement.
   */

   if (StrCompare(Keyword, Replacement, NULL, STR_MATCHLENGTH|STR_CASE) IS ERR_Okay) {
      if ((*Result = StrClone(Source))) {
         return(ERR_Okay);
      }
      else return(ERR_AllocMemory);
   }

   /*** Calculate string lengths ***/

   for (KeywordLength=0; Keyword[KeywordLength]; KeywordLength++);
   for (ReplaceLength=0; Replacement[ReplaceLength]; ReplaceLength++);

   Original  = Source;
   NewString = NULL;
   Alloc     = FALSE;
   pos = 0;
   while ((offset = StrSearch(Keyword, Source + pos, CaseSensitive)) != -1) {
      pos += offset;
      NewSize = StrLength(Source) - KeywordLength + ReplaceLength + 1;
      if (AllocMemory(NewSize, MEM_STRING|MEM_NOCLEAR, (APTR *)&NewString, NULL) IS ERR_Okay) {

      	 /*** Copy first set of bytes up to the keyword ***/

       	 for (i=0; i < pos; i++) NewString[i] = Source[i];

      	 /*** Copy the replacement ***/

      	 for (j=0; j < ReplaceLength; j++) NewString[i + j] = Replacement[j];

      	 /*** Copy the remaining bytes ***/

         for (i=0; Source[pos + KeywordLength + i] != NULL; i++) {
            NewString[pos + ReplaceLength + i] = Source[pos + KeywordLength + i];
         }
         NewString[pos + ReplaceLength + i] = 0;

         if (Alloc IS TRUE) FreeMemory(Source);
         Source = NewString;
         Alloc = TRUE;
         pos += ReplaceLength;
      }
      else break;
   }

   if (Source IS Original) { *Result = NULL; return(ERR_Search); }
   else *Result = Source;

   return(ERR_Okay);
}

/*****************************************************************************

###FUNCTION###
Name:     StrSearch()
Short:    Searches a string for a particular keyword/phrase.
Synopsis: LONG StrSearch(STRING Keyword, STRING String, LONG CaseSensitive)

###DESCRIPTION###
<p>This function allows you to search for a particular Keyword or phrase
inside a String.  You may search on a case sensitive or case insensitive
basis.</p>

###INPUT###
Search:        A string that specifies the keyword/phrase you are searching for.
String:        The string data that you wish to search.
CaseSensitive: Set to TRUE if the search should be case sensitive, otherwise FALSE.

###RESULT###
<p>Returns the byte position of the first occurance of the Keyword within the
String (possible values start from position 0).  If the Keyword could not be
found, this function returns a value of -1.</p>

###END###

*****************************************************************************/

static LONG StrSearch(STRING Keyword, STRING String, LONG CaseSensitive)
{
   LONG pos, i;

   if ((!String) OR (!Keyword)) {
      SendError(ERH_StrSearch, ERR_Args);
      return(-1);
   }

   pos = 0;
   if (CaseSensitive IS FALSE) {
      while (String[pos]) {
         for (i=0; Keyword[i]; i++) if (UCase(String[pos+i]) != UCase(Keyword[i])) break;
         if (!Keyword[i]) return(pos);
         for (++pos; (String[pos] & 0xc0) IS 0x80; pos++);
      }
   }
   else {
      while (String[pos]) {
         for (i=0; Keyword[i]; i++) if (String[pos+i] != Keyword[i]) break;
         if (!Keyword[i]) return(pos);
         for (++pos; (String[pos] & 0xc0) IS 0x80; pos++);
      }
   }

   return(-1);
}

/*****************************************************************************

###FUNCTION###
Name:     StrShrink()
Short:    Shrinks strings by destroying data.
Synopsis: void StrShrink(STRING String, LONG Position, LONG TotalBytes)

###DESCRIPTION###
<p>This function allows you to shrink the size of a string by destroying
some if its data.  You can inform the function where exactly you want to
start destroying characters and how many characters you want to delete.
For example, StrShrink(&quot;Hello World&quot;, 4, 5) gives
&quot;Helrld&quot; by deleting the characters &quot;lo Wo&quot;.</p>

<p>This function operates entirely upon the String that you pass to it, so
no memory will be allocated or returned to you.  The String will be null
terminated and its new length will be returned as a result.</p>

###INPUT###
String:   Pointer to the string that you want to shrink.
Position: The byte position that you want to start shrinking the string from.
          The position must be shorter than the full length of the string.
TotalBytes: The total number of bytes that you want to eliminate.

###RESULT###
<p>Returns the new length of the shrunken string, not including the null
byte.</p>

###END###

*****************************************************************************/

static LONG StrShrink(STRING String, LONG Position, LONG AmtChars)
{
   STRING OrigString;

   if ((String) AND (Position >= 0) AND (AmtChars > 0)) {
      OrigString = String;
      String += Position;
      while (String[AmtChars] != NULL) {
         *String = String[AmtChars];
         String++;
         AmtChars++;
      }
      String = NULL; /* Terminate the string */
   }
   else { SendError(ERH_Strings, ERR_Args); return(NULL); }

   return((LONG)(String - OrigString));
}

/*****************************************************************************

###FUNCTION###
Name:     StrSort()
Short:    Used to sort string arrays.
Synopsis: ERROR StrSort(STRING *List, LONG Flags)

###DESCRIPTION###
<p>This function is used to sort string arrays into alphabetical order.  You
will need to provide the list of unsorted strings in a block of string
pointers, terminated with a NULL entry.  For example:</p>

<pre>
  STRING List[] = {
     "banana",
     "apple",
     "orange",
     "kiwifruit",
     NULL
  }
</pre>

<p>The sorting routine will work within the confines of the array that you
have provided and will not allocate any memory when performing the sort.</p>

<p>Optional flags include SBF_NODUPLICATES, which strips duplicated strings
out of the array; SBF_CASE to acknowledge case differences when determining
string duplication and SBF_DESC to sort in descending order.</p>

###INPUT###
List: Must point to an array of string pointers, terminated with a NULL entry.
Flags: Optional flags may be set here.

###RESULT###
ERR_Okay: The list has been sorted.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR StrSort(STRING *List, LONG Flags)
{
   if (!List) return(ERR_Args);

#if 0
   /* Bubble sort, yuck... */

   LONG count, index, i, j, strflags;
   STRING temp, Name1, Name2;
   BYTE char1, char2;

   count = 1;
   index = 0;
   while (index != -1) {
      index = -1;
      for (i=0; List[i]; i++) {
         if (List[i+1]) {
            Name1 = List[i];
            Name2 = List[i+1];
            while ((*Name1) AND (*Name2)) {
               char1 = *Name1;
               char2 = *Name2;
               if ((char1 >= 'A') AND (char1 <= 'Z')) char1 = char1 - 'A' + 'a';
               if ((char2 >= 'A') AND (char2 <= 'Z')) char2 = char2 - 'A' + 'a';
               if (char1 > char2) {
                  temp = List[i+1];
                  List[i+1] = List[i];
                  List[i] = temp;
                  index = i;
                  break;
               }
               else if (char1 < char2) break;

               Name1++;
               Name2++;

               if ((Name1[0]) AND (!Name2[0])) {
                  temp = List[i+1];
                  List[i+1] = List[i];
                  List[i] = temp;
                  index = i;
                  break;
               }
            }
         }
      }
   }
#endif
   /* Shell sort.  Similar to bubble sort but much faster
   ** because it can copy records over larger distances.
   */

   LONG h, total, i, j, strflags;
   STRING temp;

   for (total=0; List[total]; total++);

   h = 1;
   while (h < total / 9) h = 3 * h + 1;

   if (Flags & SBF_DESC) {
      for (; h > 0; h /= 3) {
         for (i=h; i < total; i++) {
            temp = List[i];
            for (j=i; (j >= h) AND (sortcompare(List[j - h], temp) < 0); j -= h) {
               List[j] = List[j - h];
            }
            List[j] = temp;
         }
      }
   }
   else {
      for (; h > 0; h /= 3) {
         for (i=h; i < total; i++) {
            temp = List[i];
            for (j=i; (j >= h) AND (sortcompare(List[j - h], temp) > 0); j -= h) {
               List[j] = List[j - h];
            }
            List[j] = temp;
         }
      }
   }

   if (Flags & SBF_NODUPLICATES) {
      strflags = STR_MATCHLENGTH;
      if (Flags & SBF_CASE) strflags |= STR_MATCHCASE;

      for (i=1; List[i]; i++) {
         if (StrCompare(List[i-1], List[i], NULL, strflags) IS ERR_Okay) {
            for (j=i; List[j]; j++) List[j] = List[j+1];
            i--;
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FUNCTION###
Name:     StrToFloat()
Short:    Converts strings to floating point numbers.
Synopsis: DOUBLE StrToFloat(STRING String)

###DESCRIPTION###
<p>This function converts strings into 64-bit floating point numbers.  It
supports negative numbers (if a minus sign is at the front) and skips leading
spaces and non-numeric characters that occur before any digits.</p>

<p>If the function encounters a non-numeric character once it has started its
number crunching, it immediately stops and returns the value that has been
calculated up to that point.</p>

###INPUT###
String: Pointer to the string that is to be converted to a floating point number.

###RESULT###
<p>Returns the floating point value that was calculated from the String.</p>

###END###

*****************************************************************************/

static DOUBLE StrToFloat(STRING String)
{
   DOUBLE Result;
   LONG Factor, Number;

   if (!String) return(0);

   /*** Ignore any leading characters ***/

   while ((*String != '-') AND ((*String < '0') OR (*String > '9'))) {
      if (*String IS 0) return(0);
      String++;
   }

   Result = (DOUBLE)StrToInt(String);

   /*** Skip numbers ***/
   if (*String IS '-') String++;
   while ((*String >= '0') AND (*String <= '9')) String++;

   /*** Check for decimal place ***/
   if (*String IS '.') {
      String++;
      if ((*String >= '0') AND (*String <= '9')) {
         if ((Number = StrToInt(String))) {
             Factor = 1;
             while ((*String >= '0') AND (*String <= '9')) {
                Factor = Factor * 10;
                String++;
             }
             if (Result < 0) {
                Result = Result - (((DOUBLE)Number) / Factor);
             }
             else Result = Result + (((DOUBLE)Number) / Factor);
         }
      }
   }

   return(Result);
}

/*****************************************************************************

###FUNCTION###
Name:     StrToInt()
Short:    Converts a string to an integer.
Synopsis: LARGE StrToInt(STRING String)

###DESCRIPTION###
<p>This function converts a String to its integer equivalent.  It supports
negative numbers (if a minus sign is at the front) and skips leading spaces
and non-numeric characters that occur before any digits.</p>

<p>If the function encounters a non-numeric character once it has started its
digit processing, it immediately stops and returns the result calculated
up to that point.</p>

<p>Here are some string conversion examples:</p>

<pre>
   <b>String</b>         <b>Result</b>
   183        =   183
     2902a6   =   2902
   hx239      =   239
   -45        =   -45
    jff-9     =   -9
</pre>

###INPUT###
String: Pointer to the string that is to be converted to an integer.

###RESULT###
<p>Returns the integer value that was calculated from the String.</p>

###END###

*****************************************************************************/

static LARGE StrToInt(STRING str)
{
  DOUBLE number;
  BYTE neg;

  if (str) {

     /*** Ignore any leading characters ***/

     neg = FALSE;
     while ((*str < '0') OR (*str > '9')) {
        if (*str IS 0) return(NULL);
        if (*str IS '-') neg = TRUE;
        if (*str IS '+') neg = FALSE;
        str++;
     }

     /*** Ignore leading zeros ***/

     while (*str IS '0') str++;

     /*** Check if there are any numbers following the leading zeros ***/

     number = 0;
     if ((*str >= '1') AND (*str <= '9')) {
        while (*str) {
           if ((*str >= '0') AND (*str <= '9')) {
              number *= 10;
              if (neg IS TRUE) {
                 number -= (*str - '0');
              }
              else number += (*str - '0');
           }
           else return(number);
           str++;
        }

        return(number);
     }
     else return(0);
  }
  else return(NULL);
}

/*****************************************************************************

###FUNCTION###
Name:     StrTranslate()
Short:    Translates object references that have been declared within strings.
Synopsis: ERROR StrTranslate(STRING Buffer, LONG Length, LONG Flags)

###DESCRIPTION###
<p>This function is used to translate strings that make object and field
references using the standard referencing format.  References are made to
objects by enclosing statements within square brackets.  As a result of
calling this function, all references within the Buffer will be translated to
their relevant format.  The Buffer needs to be large enough to accommodate
these adjustments as it will be expanded during the translation.  It is
recommended that the Buffer is at least two times the actual length of the
string that you are translating.</p>

<p>Valid references can be made to an object by name, ID or relative
parameters.  Here are some examples illustrating the different variations:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Reference</td><td class="tableheader" bgcolor="#eeeeee">Type</td></tr>
<tr><td width="10%">[render]</td><td>Name reference.</td></tr>
<tr><td width="10%">[#49302]</td><td>ID reference.</td></tr>
<tr><td width="10%">[self]</td><td>Relative reference to the object that has the current context.</td></tr>
<tr><td width="10%">[owner]&nbsp;&nbsp;</td><td>Relative reference to the current object's owner.</td></tr>
<tr><td width="10%">[task]</td><td>Refers to the current task ID.</td></tr>
</table>

<p>Object references are always converted to their ID equivalent when
StrTranslate() is called, so the [render], [self], and [owner]
statements will all be converted to their associated ID numbers - for
example [#96043].  ID references that already exist in a string prior to
translation will not be affected in any way.</p>

<p>Field references are a slightly different matter and will be converted to
the value of the field that they are referencing.  A field reference is
defined using the object referencing format, but they contain a '.fieldname'
extension.  Here are some examples:</p>

<pre>
   [render.width]
   [file.location]
</pre>

<p>A string such as &quot;[mywindow.height] + [mywindow.width]&quot; could be
translated to &quot;255 + 120&quot; for instance.  References to string based
fields can expand the Buffer very quickly, which is why large buffer spaces
are recommended for all-purpose translations.</p>

###INPUT###
Buffer: The buffer that you wish to translate.
Length: The total size of the Buffer.
Flags:  Currently unused, set to NULL.

###RESULT###
ERR_Okay:    Translation successful.
ERR_Args:    Invalid arguments were specified.
ERR_BadData: The buffer does not contain any object references.
ERR_Failed:  One of the references is invalid or does not refer to an actual object.

###END###

*****************************************************************************/

static ERROR StrTranslate(STRING Buffer, LONG BufferLength, LONG Flags)
{
   OBJECTID ObjectID;
   LONG pos, i, j;
   OBJECTPTR ObjectPtr;
   WORD endbracket;
   ERROR error;
   UBYTE Name[MAX_NAMELEN];
   UBYTE Field[60];
   UBYTE Value[300];

   if (!Buffer) return(SendError(ERH_Strings, ERR_Args));

   if (BufferLength <= 0) BufferLength = 0x7fff;

   /*** Quick check for translation symbols ***/

   for (i=0; Buffer[i] != '['; i++) {
      if (!Buffer[i]) return(ERR_BadData);
   }

   error = ERR_Okay;

   /*** Skip to the end of the buffer (translation occurs 'backwards') ***/

   for (pos=0; Buffer[pos]; pos++);
   pos--;
   while (pos >= 0) {
      /*** Do not translate quoted areas ***/

      if ((Buffer[pos] IS '"') AND (!(Flags & STF_IGNOREQUOTES))) {
         pos--;
         while ((pos >= 0) AND (Buffer[pos] != '"')) pos--;
         if (pos < 0) {
            DPrintF("@StrTranslate:","Badly defined string: %.60s", Buffer);
            return(ERR_BadData);
         }
      }

      if (Buffer[pos] IS '\\') {
         if (Buffer[pos+1] IS '\\') {
            Buffer[pos] = '\\';
         }
         else if (Buffer[pos+1] IS '[') {
            Buffer[pos] = '[';
         }

         for (i=pos+1, j=pos; Buffer[i]; i++) Buffer[j++] = Buffer[i];
         Buffer[j] = 0;
         pos--;
      }
      else if ((Buffer[pos] IS '[') AND ((pos < 1) OR (Buffer[pos-1] != '\\'))) {
         /*** Make sure that there is a closing bracket ***/
         for (endbracket=pos; (Buffer[endbracket] != ']') AND (Buffer[endbracket]); endbracket++);

         if (Buffer[endbracket] != ']') {
            DPrintF("@StrTranslate:","Badly defined string: %.60s ...", Buffer);
            return(ERR_BadData);
         }

         /*** Retrieve the name of the object ***/

         j = 0;
         for (i=pos+1; (Buffer[i] != '.') AND (Buffer[i] != ']'); i++) {
            if (j < MAX_NAMELEN-1) Name[j++] = Buffer[i];
         }
         Name[j] = 0;

         /*** Get the object ID ***/

         if (StrCompare(Name, "self", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            ObjectID = GetUniqueID(CurrentContext());
         }
         else if (StrCompare(Name, "container", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            ObjectID = GetOwner(CurrentContext());
         }
         else if (StrCompare(Name, "owner", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            ObjectID = GetOwner(CurrentContext());
         }
         else if (StrCompare(Name, "task", NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            ObjectID = CurrentTaskID();
         }
         else if (*Name) {
            if (FastFindObject(Name, NULL, &ObjectID, 1, NULL) != ERR_Okay) ObjectID = NULL;
         }
         else ObjectID = NULL;


         if (ObjectID) {
            if (Buffer[i] IS '.') {
               /*** Get the field from the object ***/
               i++;
               if (AccessObject(ObjectID, 1000, &ObjectPtr) IS ERR_Okay) {
                  j = 0;
                  while ((i < endbracket) AND (j < 59)) {
                     Field[j++] = Buffer[i++];
                  }
                  Field[j] = 0;

                  if ((error = GetFieldVariable(ObjectPtr, Field, Value, sizeof(Value))) IS ERR_Okay) {
                     error = InsertString(Value, Buffer, pos, endbracket-pos+1, BufferLength);
                     if (error) {
                        DPrintF("@StrTranslate:","Error inserting string \"%s\" into position %d of %d byte buffer.", Value, pos, BufferLength);
                     }
                  }

                  ReleaseObject(ObjectPtr);
               }
            }
            else {
               /*** Convert the object reference to an ID ***/
               Value[0] = '#';
               IntToStr(ObjectID, Value+1, 30);
               error = InsertString(Value, Buffer, pos, endbracket-pos+1, BufferLength);
               if (error) {
                  DPrintF("@StrTranslate:","Error inserting object reference \"%s\".", Value);
               }
            }
         }
         else error = ERR_Failed;

         if (error != ERR_Okay) {
            if (Flags & STF_KEEPINTEGRITY) {
               pos--;
            }
            else {
               /* If an error occurred, delete everything contained by
               ** the square brackets to prevent recursion errors.
               */

               for (i=endbracket+1; Buffer[i]; i++) {
                  Buffer[pos++] = Buffer[i];
               }
               Buffer[pos] = 0;
               error = ERR_Okay;
            }
         }
      }
      else pos--;
   }

   return(error);
};

/*****************************************************************************

###FUNCTION###
Name:     StrUpper()
Short:    Changes a string so that all alpha characters are in upper-case.
Synopsis: void StrUpper(STRING String)

###DESCRIPTION###
<p>This function will alter a String so that all upper-case characters are
changed to lower-case.  Non lower-case characters are unaffected by this
function.  Here is an example:</p>

<pre>
   "HeLLo world" = "HELLO WORD"
</pre>

###INPUT###
String: Pointer to the string that you want in upper-case.

###END###

*****************************************************************************/

static void StrUpper(STRING String)
{
   if (!String) return;

   while (*String) {
      if ((*String >= 'a') AND (*String <= 'z')) *String -= 0x20;
      String++;
   }
}

/*****************************************************************************

###FUNCTION###
Name:     UTF8CharacterOffset()
Short:    Retrieves the byte position of a character in a UTF-8 string.
Synopsis: LONG UTF8CharacterOffset(STRING String, LONG Position)

###DESCRIPTION###
<p>This function is used for determing the actual byte position of a character
within a UTF-8 string.  It should be used in situations where you need to know
the precise location of a character within a UTF-8 string.</p>

###INPUT###
String:   A null-terminated UTF-8 string.
Position: The character number that you need a byte offset for.

###RESULT###
<p>Returns the byte offset of the character.</p>

###END###

*****************************************************************************/

static LONG UTF8CharacterOffset(STRING String, LONG Index)
{
   LONG offset;

   if (!String) return(0);

   offset = 0;
   while ((String[offset]) AND (Index > 0)) {
      for (++offset; ((String[offset] & 0xc0) IS 0x80); offset++);
      Index--;
   }
   return(offset);
}

/*****************************************************************************

###FUNCTION###
Name:     UTF8CharacterLength()
Short:    Returns the number of bytes used to define a single UTF-8 character.
Synopsis: LONG UTF8CharacterLength(STRING String)

###DESCRIPTION###
<p>This function will return the total number of bytes used to create
a single UTF-8 character.  You are only required to supply a pointer to the
start of the character string that you want to analyse.</p>

###INPUT###
String: Pointer to a UTF-8 string.

###RESULT###
<p>Returns the number of bytes used to create the UTF-8 character referred to
by the String argument.</p>

###SEE ALSO###
Function: UTF8CharacterLength()

###END###

*****************************************************************************/

static LONG UTF8CharacterLength(STRING String)
{
   LONG total;
   if ((!String) OR (!*String)) return(0);
   for (total=1; ((String[total] & 0xc0) IS 0x80); total++);
   return(total);
}

/*****************************************************************************

###FUNCTION###
Name:     UTF8Length()
Short:    Returns the total number of characters in a UTF-8 string.
Synopsis: LONG UTF8Length(STRING String)

###DESCRIPTION###
<p>This function will return the 'real' length of a UTF-8 string.  Using
the StrLength() will return the number of bytes in a UTF-8 string, but
to find out the actual number of characters, you must use UTF8Length().</p>

###INPUT###
String: Pointer to a UTF-8 string.

###RESULT###
<p>Returns the total number of characters used in the supplied UTF-8
string.</p>

###SEE ALSO###
Function: UTF8CharacterLength()

###END###

*****************************************************************************/

static LONG UTF8Length(STRING String)
{
   LONG total;

   if (!String) return(0);

   for (total=0; *String; total++) {
      for (++String; ((*String & 0xc0) IS 0x80); String++);
   }
   return(total);
}

/*****************************************************************************

###FUNCTION###
Name:     UTF8OffsetToCharacter()
Short:    Converts a byte offset into a character position.
Synopsis: LONG UTF8OffsetToCharacter(STRING String, LONG Position)

###DESCRIPTION###
<p>This function is used for determing the character number of a byte position
within a UTF-8 string.  It should be used in situations where you need to know
the precise location of a character within a UTF-8 string.</p>

###INPUT###
String:   A null-terminated UTF-8 string.
Position: The byte offset that you need a character number for.

###RESULT###
<p>Returns the number of the character at the given byte position.</p>

###END###

*****************************************************************************/

static LONG UTF8OffsetToCharacter(STRING String, LONG Offset)
{
   LONG pos;

   if (!String) return(0);

   pos = 0;
   while ((Offset) AND (String[pos])) {
      for (++pos; ((String[pos] & 0xc0) IS 0x80); pos++);
      Offset--;
   }
   return(pos);
}

/*****************************************************************************

###FUNCTION###
Name:     UTF8PrevLength()
Short:    Gets the byte length of the previous character at a specific position in a UTF-8 string.
Synopsis: LONG UTF8PrevLength(STRING String, LONG Index)

###DESCRIPTION###
<p>If you need to get the byte-length of the previous character from within
a UTF-8 string, you can use this function.  You need to provide the byte Index
from which you want to calculate the previous character in the text as specified
by String.</p>

###INPUT###
String: Pointer to a UTF-8 string.
Index:  The byte index from which the size of the previous character should be calculated.

###RESULT###
<p>Returns the byte-length of the previous character.</p>

###END###

*****************************************************************************/

static LONG UTF8PrevLength(STRING String, LONG ByteIndex)
{
   LONG len;
   len = 0;
   for (--ByteIndex; ByteIndex > 0; --ByteIndex) {
      len++;
      if ((String[ByteIndex] & 0xc0) != 0x80) return(len);
   }
   return(len);
}

/*****************************************************************************

###FUNCTION###
Name:     UTF8ReadValue()
Short:    Converts UTF-8 characters into 32-bit unicode values.
Synopsis: LONG UTF8ReadValue(STRING String, LONG *Length)

###DESCRIPTION###
<p>This function is used to convert UTF-8 characters into 32-bit unicode
values.  Simply provide it with the address of the UTF-8 character that
you need to convert and it will return the calculated unicode value.  If you
need to know how many bytes where used in making up the UTF-8 character,
set the Length argument so that the function can return the total number
of bytes to you.</p>

###INPUT###
String: Pointer to a character in a UTF-8 string that you want to convert.
Length: Optional argument that will store the resulting number of bytes that make up the UTF-8 character.

###RESULT###
<p>Returns the extracted unicode value.</p>

###END###

*****************************************************************************/

static LONG UTF8ReadValue(STRING String, LONG *Length)
{
   WORD i;
   ULONG code;

   if (*String IS 0) {
      if (Length) *Length = 0;
      return(0);
   }
   else if (*String < 128) {
      if (Length) *Length = 1;
      return(*String);
   }
   else if ((*String & 0xe0) IS 0xc0) {
      if (Length) *Length = 2;
      return(((String[0] & 0x1f)<<6) | (String[1] & 0x3f));
   }
   else if ((*String & 0xf0) IS 0xe0) {
      if (Length) *Length = 3;
      code = *String & 0x0f;
      for (i=1; i < 3; i++) {
         if ((String[i] & 0xc0) != 0x80) return(0);
         code = (code<<6) | (String[i] & 0x3f);
      }
   }
   else if ((*String & 0xf8) IS 0xf0) {
      if (Length) *Length = 4;
      code = *String & 0x07;
      for (i=1; i < 4; i++) {
         if ((String[i] & 0xc0) != 0x80) return(0);
         code = (code<<6) | (String[i] & 0x3f);
      }
   }
   else if ((*String & 0xfc) IS 0xf8) {
      if (Length) *Length = 5;
      code = *String & 0x03;
      for (i=1; i < 5; i++) {
         if ((String[i] & 0xc0) != 0x80) return(0);
         code = (code<<6) | (String[i] & 0x3f);
      }
   }
   else if ((*String & 0xfc) IS 0xfc) {
      if (Length) *Length = 6;
      code = *String & 0x01;
      for (i=1; i < 6; i++) {
         if ((String[i] & 0xc0) != 0x80) return(0);
         code = (code<<6) | (String[i] & 0x3f);
      }
   }
   else {
      if (Length) *Length = 1;
      return(0);
   }

   return(code);
}

/*****************************************************************************

###FUNCTION###
Name:     UTF8WriteValue()
Short:    Writes a 32-bit unicode value into a UTF-8 character buffer.
Synopsis: LONG UTF8WriteValue(LONG Value, STRING String, LONG StringSize)

###DESCRIPTION###
<p>This function is used to write out unicode values in UTF-8 string format.
You need to provide a string buffer, keeping in mind that anything from 1
to 6 bytes may be written to the buffer, depending on the size of the unicode
value.</p>

<p>If the buffer is not large enough to hold the UTF-8 string then the
function will do nothing and return a value of zero.  Otherwise, the
character will be written and the total number of bytes used will be
returned.</p>

<p>This function will not add a null terminator to the end of the UTF-8
string.</p>

###INPUT###
Value:  The 32-bit unicode value that you want to write.
String: Pointer to a string buffer that will hold the UTF-8 characters.
StringSize: The size of the destination string buffer (does not need to be any larger than 6 bytes).

###RESULT###
<p>Returns the total amount of characters written to the string buffer.</p>

###END###

*****************************************************************************/

static LONG UTF8WriteValue(LONG Value, STRING String, LONG StringSize)
{
   if (Value < 128) {
      if (Value < 0) return(0);
      *String = (UBYTE)Value;
      return(1);
   }
   else if (Value < 0x800) {
      if (StringSize < 2) return(0);
      String[1] = (Value & 0x3f) | 0x80;
      Value  = Value>>6;
      String[0] = Value | 0xc0;
      return(2);
   }
   else if (Value < 0x10000) {
      if (StringSize < 3) return(0);
      String[2] = (Value & 0x3f)|0x80;
      Value  = Value>>6;
      String[1] = (Value & 0x3f)|0x80;
      Value  = Value>>6;
      String[0] = Value | 0xe0;
      return(3);
   }
   else if (Value < 0x200000) {
      if (StringSize < 4) return(0);
      String[3] = (Value & 0x3f)|0x80;
      Value  = Value>>6;
      String[2] = (Value & 0x3f)|0x80;
      Value  = Value>>6;
      String[1] = (Value & 0x3f)|0x80;
      Value  = Value>>6;
      String[0] = Value | 0xf0;
      return(4);
   }
   else if (Value < 0x4000000) {
      if (StringSize < 5) return(0);
      String[4]  = (Value & 0x3f)|0x80;
      Value = Value>>6;
      String[3]  = (Value & 0x3f)|0x80;
      Value = Value>>6;
      String[2]  = (Value & 0x3f)|0x80;
      Value = Value>>6;
      String[1]  = (Value & 0x3f)|0x80;
      Value = Value>>6;
      String[0]  = Value | 0xf8;
      return(5);
   }
   else {
      if (StringSize < 6) return(0);
      String[5]  = (Value & 0x3f)|0x80;
      Value = Value>>6;
      String[4]  = (Value & 0x3f)|0x80;
      Value = Value>>6;
      String[3]  = (Value & 0x3f)|0x80;
      Value = Value>>6;
      String[2]  = (Value & 0x3f)|0x80;
      Value = Value>>6;
      String[1]  = (Value & 0x3f)|0x80;
      Value = Value>>6;
      String[0]  = Value | 0xfc;
      return(6);
   }
}

/*****************************************************************************
** Internal: InsertString()
*/

static ERROR InsertString(STRING Insert, STRING Buffer, LONG Pos, LONG ReplaceChars,
                   LONG BufferLength)
{
   LONG InsertLen, i, StringLen, j;

   for (InsertLen=0; Insert[InsertLen]; InsertLen++);

   /*** String insertion ***/

   if (InsertLen < ReplaceChars) {
      i = Pos + StrCopy(Insert, Buffer+Pos, COPY_ALL);
      i = Pos + ReplaceChars;
      Pos += InsertLen;
      while (Buffer[i]) Buffer[Pos++] = Buffer[i++];
      Buffer[Pos] = NULL;
   }
   else if (InsertLen IS ReplaceChars) {
      while (*Insert) Buffer[Pos++] = *Insert++;
   }
   else {

      /*** Check if an overflow will occur ***/

      for (StringLen=0; Buffer[StringLen]; StringLen++);
      if ((BufferLength - 1) < (StringLen + (ReplaceChars - InsertLen))) {
         DPrintF("@InsertString()","Insert: \"%s\", Buffer: \"%s\", [Pos %d], No: %d, Buffer Size: %d", Insert, Buffer + Pos, Pos, ReplaceChars, BufferLength);
         DPrintF("@InsertString:","Buffer overflow (%d bytes) while translating line \"%.60s\"", BufferLength, Buffer);
         return(ERR_BufferOverflow);
      }

      /*** Expand the string ***/
      i = StringLen + (InsertLen - ReplaceChars) + 1;
      StringLen += 1;
      j = StringLen-Pos-ReplaceChars+1;
      while (j > 0) {
         Buffer[i--] = Buffer[StringLen--];
         j--;
      }

      /*** Copy the insert string into the position ***/
      for (i=0; Insert[i]; i++, Pos++) Buffer[Pos] = Insert[i];
   }

   return(ERR_Okay);
}

